unit FileSplitter;

{---------------------------------------------------------------------------}
{  TFileSplitter v1.00                                                      }
{---------------------------------------------------------------------------}
{***************************************************************************}
{                                                                           }
{       Component description:                                              }
{                                                                           }
{       TFileSplitter is an invisible component with a simple task to       }
{  split a file into multiple files of certain size. The most common need   }
{  for this is when you want to put your (big) file on floppy disks, for    }
{  backup or any other purposes. This component also includes method for    }
{  reversing the process - the unsplitting. All existing files will be      }
{  overwritten without warning during the process of splitting/unsplitting. }
{       When splitting, if the SplitFileName property is empty, a file(s)   }
{  with the same name as FileName will be created, with extension(s) as     }
{  follows: '.SPL', '.001', '.002', etc.                                    }
{       When unsplitting, if the FileName property is empty, a file with    }
{  the same name as SplitFileName will be created, with extension '.XXX'.   }
{                                                                           }
{***************************************************************************}
{                                                                           }
{       Properties:                                                         }
{       -----------                                                         }
{       BufferSize     - Size of buffer used for reading input data and     }
{                        writing to output file(s).                         }
{       FileName       - The name of a file to be split, or the name of a   }
{                        resulting file in case of unsplitting.             }
{       Size           - Maximum size (in bytes) of the resulting           }
{                        (splitted) file(s). Not needed in case of          }
{                        unsplitting.                                       }
{       SplitFileName  - The name of a resulting file, or the name of the   }
{                        first file in case of unsplitting.                 }
{                                                                           }
{       Events:                                                             }
{       -------                                                             }
{       OnNeedDisk     - If assigned, this event is fired before the        }
{                        creation of each splitted file. Also, this event   }
{                        is fired before appending each splitted file to    }
{                        the original file, in case of unsplitting.         }
{                                                                           }
{       Methods:                                                            }
{       --------                                                            }
{       Split          - This method executes the splitting.                }
{       UnSplit        - This method does the unsplitting.                  }
{                                                                           }
{***************************************************************************}
{                                                                           }
{       Known bugs:                                                         }
{                                                                           }
{       There are no known bugs at this time.                               }
{                                                                           }
{***************************************************************************}
{                                                                           }
{       Version history:                                                    }
{                                                                           }
{       March 08, 2000       - v1.00       - Initial release                }
{                                                                           }
{***************************************************************************}
{                                                                           }
{          Copyright (c) 2000 by Jovan Sedlan. All rights reserved.         }
{                                                                           }
{  Copyright:                                                               }
{                                                                           }
{  TFileSplitter (hereafter "component") source code, and any other source  }
{  code inside this archive is copyrighted by Jovan Sedlan (hereafter       }
{  "author"), and shall remain the exclusive property of the author.        }
{                                                                           }
{  Distribution Rights:                                                     }
{                                                                           }
{  You are granted a non-exlusive, royalty-free right to produce and        }
{  distribute compiled binary files (executables, DLLs, etc.) that are      }
{  built with this component's source code unless specifically stated       }
{  otherwise.                                                               }
{  You are further granted permission to redistribute the component source  }
{  code in source code form, provided that the original archive as found on }
{  the DSP web site (http://delphi.icm.edu.pl/) is distributed unmodified.  }
{  For example, if you create a descendant of TFileSplitter, you must       }
{  include in  the distribution package the FileSpl.zip file in the exact   }
{  form that you downloaded it from http://delphi.icm.edu.pl/               }
{                                                                           }
{  Restrictions:                                                            }
{                                                                           }
{  Without the express written consent of the author, you may not:          }
{   * Distribute modified versions of any DFS source code by itself. You    }
{     must include the original archive as you found it on the web.         }
{   * Sell or lease any portion of DFS source code. You are, of course,     }
{     free to sell any of your own original code that works with, enhances, }
{     etc. this component's source code.                                    }
{   * Distribute component's source code for profit.                        }
{                                                                           }
{  Warranty:                                                                }
{                                                                           }
{  There is absolutely no warranty of any kind whatsoever with this         }
{  component's source code (hereafter "software"). The software is provided }
{  to you "AS-IS", and all risks and losses associated with it's use are    }
{  assumed by you. In no event shall the author of the softare, Jovan       }
{  Sedlan, be held accountable for any damages or losses that may occur     }
{  from use or misuse of the software.                                      }
{                                                                           }
{  Support:                                                                 }
{                                                                           }
{  Support is provided via email only. The source code for this software is }
{  provided free of charge. As such, I can not guarantee any support        }
{  whatsoever. While I do try to answer all questions that I receive, and   }
{  address all problems that are reported to me, you must understand that I }
{  simply can not guarantee that this will always be so.                    }
{                                                                           }
{---------------------------------------------------------------------------}
{  The latest version of my components are always available on the web at:  }
{    http://delphi.icm.edu.pl/                                              }
{---------------------------------------------------------------------------}
{  Date last modified:  March 8, 2000.                                      }
{---------------------------------------------------------------------------}
{***************************************************************************}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DsgnIntf;

type
  EFileError = class(Exception);

  TNeedDiskEvent = procedure(Sender : TObject; DiskID: word; var Continue: Boolean) of object;

  TFSFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TFSSplitFilenameProperty = class(TStringProperty)
  public
    procedure Edit; override;
    function GetAttributes: TPropertyAttributes; override;
  end;

  TFileSplitter = class(TComponent)
  private
    FFileName: TFileName;
    FSize: LongInt;
    FSplitFileName: TFileName;
    FOutDifferent: Boolean;
    FFile: TFileStream;
    FSplitFile: TFileStream;
    FBuffer: Pointer;
    FBufferSize: Integer;

    FOnNeedDisk: TNeedDiskEvent;

    procedure SetFileName(Value: TFileName);
    procedure SetSize(Value: LongInt);
    procedure SetBufferSize(Value: Integer);
    procedure SetSplitFileName(Value: TFileName);
  protected
  public
    constructor Create(AOwner: TComponent); override;
    procedure Split;
    procedure UnSplit;
  published
    property FileName: TFileName read FFileName write SetFileName;
    property SplitFileName: TFileName read FSplitFileName write SetSplitFileName;
    property Size: LongInt read FSize write SetSize;
    property BufferSize: Integer read FBufferSize write SetBufferSize;

    property OnNeedDisk: TNeedDiskEvent read FOnNeedDisk write FOnNeedDisk;
  end;

procedure Register;

implementation

{ TFSFilenameProperty }

procedure TFSFilenameProperty.Edit;
var
  FSFileOpen: TOpenDialog;
  
begin
  FSFileOpen := TOpenDialog.Create(Application);
  FSFileOpen.Filename := GetValue;
  FSFileOpen.Filter := 'All files (*.*)|*.*';
  FSFileOpen.Options := FSFileOpen.Options + [ofShowHelp, ofPathMustExist];
  FSFileOpen.Title := 'Select a file to split';
  try
    if FSFileOpen.Execute then
      SetValue(FSFileOpen.Filename);
  finally
    FSFileOpen.Free;
  end;
end;

function TFSFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

{ TFSSplitFilenameProperty }

procedure TFSSplitFilenameProperty.Edit;
var
  FSFileSave: TSaveDialog;

begin
  FSFileSave := TSaveDialog.Create(Application);
  FSFileSave.Filename := GetValue;
  FSFileSave.Filter := 'Split files (*.spl)|*.SPL';
  FSFileSave.Options := FSFileSave.Options + [ofShowHelp, ofPathMustExist];
  FSFileSave.Title := 'Select the first split file (.SPL)';
  try
    if FSFileSave.Execute then
      SetValue(FSFileSave.Filename);
  finally
    FSFileSave.Free;
  end;
end;

function TFSSplitFilenameProperty.GetAttributes: TPropertyAttributes;
begin
  Result := [paDialog, paRevertable];
end;

{ TFileSplitter }

{ Sets the name of the file to split }
procedure TFileSplitter.SetFileName(Value: TFileName);
begin
  if Value <> FFileName then
  begin
    FFileName := Value;
    if not FOutDifferent then
      FSplitFileName := ChangeFileExt(Value, '.SPL');
  end;
end;

{ Sets the name of the first split file - all other split files will
  have the same name, with extensions .001, .002, .003 etc. }
procedure TFileSplitter.SetSplitFileName(Value: TFileName);
begin
  if Value <> FSplitFileName then
  begin
    FSplitFileName := Value;
    FOutDifferent := True;
  end;
end;

{ Sets the maximum size of split files }
procedure TFileSplitter.SetSize(Value: LongInt);
begin
  if Value <> FSize then
    FSize := Value;
end;

{ Sets the buffer size }
procedure TFileSplitter.SetBufferSize(Value: Integer);
begin
  if Value <> FBufferSize then
    FBufferSize := Value;
end;

{ Component constructor... }
constructor TFileSplitter.Create(AOwner: TComponent);
begin
  inherited;
  FFileName := '';
  FSplitFileName := FFileName;
  FOutDifferent := False;
  FSize := 1457664;
  FBufferSize := 1457664;
end;

{ EraseFile deletes a file on disk }
procedure EraseFile(FN: TFileName);
var
  f: File;

begin
  AssignFile(f, FN);
  Erase(f);
end;

{ StrZero returns a string representation of a Number,
  Len characters long, with leading zeroes }
function StrZero(Number: LongInt; Len: Byte): String;
var
  Temp: String;

begin
  Temp := Trim(IntToStr(Number));
  Result := StringOfChar('0', Len - Length(Temp)) + Temp;
end;

{ Split method does the actual work - splits a file into multiple files }
procedure TFileSplitter.Split;
var
  i: Integer;
  Continue: Boolean;
  FileCount: Integer;
  sFileName: TFileName;
  BytesToRead, BytesWritten, BytesRead, BytesStored: LongInt;

begin
  if FileExists(FFileName) then
  begin
    { Make sure the filename is correct }
    if FSplitFileName = '' then
      FSplitFileName := ChangeFileExt(FFileName, '.SPL');

    { Prepare for reading from file }
    FFile := TFileStream.Create(FFileName, fmOpenRead);
    GetMem(FBuffer, FBufferSize);

    { Calculate the number of resulting files }
    FileCount := FFile.Size div FSize;
    if (FFile.Size mod FSize) <> 0 then
      Inc(FileCount);

    try
      { Write resulting file(s) }
      i := 1;
      Continue := True;
      if Assigned(FOnNeedDisk) then
        FOnNeedDisk(Self, i, Continue);
      sFileName := FSplitFileName;
      while (i <= FileCount) and Continue do
      begin
        { Delete the resulting file if it exists }
        if FileExists(sFileName) then
          EraseFile(sFileName);
        FSplitFile := TFileStream.Create(sFileName, fmCreate);

        { Write one file }
        BytesToRead := FBufferSize;
        if BytesToRead > FSize then
          BytesToRead := FSize;
        BytesWritten := 0;
        while BytesWritten < FSize do
        begin
          if BytesToRead > FSize - BytesWritten then
            BytesToRead := FSize - BytesWritten;
          BytesRead := FFile.Read(FBuffer^, BytesToRead);
          BytesStored := FSplitFile.Write(FBuffer^, BytesRead);
          Inc(BytesWritten, BytesStored);
          if FFile.Position = FFile.Size then
            Break;
        end;
        FSplitFile.Free;
        sFileName := ChangeFileExt(sFileName, '.' + StrZero(i, 3));
        Inc(i);
        if Assigned(FOnNeedDisk) then
          FOnNeedDisk(Self, i, Continue);
      end;
    finally
      { Clean up }
      FreeMem(FBuffer);
      FFile.Free;
    end;
  end
  else
    raise EFileError.Create('Input file ''' + FFileName + ''' does not exist.');
end;

{ Method UnSplit creates the original file from the split files. Name and
  extension of the original file must be set in FileName property }
procedure TFileSplitter.UnSplit;
var
  i: Integer;
  Continue: Boolean;
  sFileName: TFileName;
  BytesRead: LongInt;

begin
  if FileExists(FSplitFileName) then
  begin
    { Make sure the filename is correct }
    if FFileName = '' then
      FFileName := ChangeFileExt(FSplitFileName, '.XXX');

    { Delete the resulting file if it exists }
    if FileExists(FFileName) then
      EraseFile(FFileName);

    { Prepare for writing to file }
    FFile := TFileStream.Create(FFileName, fmCreate);
    GetMem(FBuffer, FBufferSize);

    try
      i := 0;
      sFileName := FSplitFileName;

      { Read files one by one and write to resulting file }
      Continue := True;
      if Assigned(FOnNeedDisk) then
        FOnNeedDisk(Self, i + 1, Continue);
      while FileExists(sFileName) and Continue do
      begin
        FSplitFile := TFileStream.Create(sFileName, fmOpenRead);
        try
          while FSplitFile.Size <> FSplitFile.Position do
          begin
            BytesRead := FSplitFile.Read(FBuffer^, FBufferSize);
            FFile.Write(FBuffer^, BytesRead);
          end;
        finally
          FSplitFile.Free;
        end;
        Inc(i);
        sFileName := ChangeFileExt(sFileName, '.' + StrZero(i, 3));
        if Assigned(FOnNeedDisk) then
          FOnNeedDisk(Self, i + 1, Continue);
      end;
    finally
      FreeMem(FBuffer);
      FFile.Free;
    end;
  end
  else
    raise EFileError.Create('Input file ''' + FSplitFileName + ''' does not exist.');
end;

{ Register component and property editors }
procedure Register;
begin
  RegisterComponents('SEDLAN A.D.', [TFileSplitter]);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileSplitter, 'FileName', TFSFileNameProperty);
  RegisterPropertyEditor(TypeInfo(TFileName), TFileSplitter, 'SplitFileName', TFSSplitFileNameProperty);
end;

end.

