unit Graphics32;

{**************************************************}
{  A Library for Fast 32-Bit Bitmap Manipulation   }
{  Copyright (c) 2000 Alex Denissov                }
{  Version: 0.9                                    }
{  Date: 09-Mar-2000                               }
{  See Licence.txt for license information         }
{**************************************************}

interface

uses
  Windows, Classes, SysUtils, Messages, Controls, Graphics;

type
  { New color types }
  TColor32 = type Longword;
  PColor32 = ^TColor32;
  TColor32Array = array [0..0] of TColor32;
  PColor32Array = ^TColor32Array;
  TPalette32 = array [Byte] of TColor32;
  PPalette32 = ^TPalette32;

  { Alignment of image and sprites in TImage32 }
  TScaleMode = (smNormal, smStretch, smScale, smResize);
  TBitmapAlign = (baTopLeft, baCenter);
  TSpriteOrigin = (soImage, soBitmap);

const
  { Some predefined color constants }
  {$WRITEABLECONST OFF}
  clBlack32               : TColor32 = $FF000000;
  clDimGray32             : TColor32 = $FF3F3F3F;
  clGray32                : TColor32 = $FF7F7F7F;
  clLightGray32           : TColor32 = $FFBFBFBF;
  clWhite32               : TColor32 = $FFFFFFFF;
  clMaroon32              : TColor32 = $FF7F0000;
  clGreen32               : TColor32 = $FF007F00;
  clOlive32               : TColor32 = $FF7F7F00;
  clNavy32                : TColor32 = $FF00007F;
  clPurple32              : TColor32 = $FF7F007F;
  clTeal32                : TColor32 = $FF007F7F;
  clRed32                 : TColor32 = $FFFF0000;
  clLime32                : TColor32 = $FF00FF00;
  clYellow32              : TColor32 = $FFFFFF00;
  clBlue32                : TColor32 = $FF0000FF;
  clFuchsia32             : TColor32 = $FFFF00FF;
  clAqua32                : TColor32 = $FF00FFFF;

  { Some semi-transparent color constants }
  clTrWhite32             : TColor32 = $7FFFFFFF;
  clTrBlack32             : TColor32 = $7F000000;
  clTrRed32               : TColor32 = $7FFF0000;
  clTrGreen32             : TColor32 = $7F00FF00;
  clTrBlue32              : TColor32 = $7F0000FF;

{ Color construction and conversion }
function Color32(WinColor: TColor): TColor32; overload;
function Color32(R, G, B: Byte; A: Byte = $FF): TColor32; overload;
function Color32(Index: Byte; var Palette: TPalette32): TColor32; overload;
function Gray32(Intensity: Byte; Alpha: Byte = $FF): TColor32;
function WinColor(Color32: TColor32): TColor;

{ Color component access }
function RedComponent(Color32: TColor32): Integer;
function GreenComponent(Color32: TColor32): Integer;
function BlueComponent(Color32: TColor32): Integer;
function AlphaComponent(Color32: TColor32): Integer;
function Intensity(Color32: TColor32): Integer;

{ Alpha adjustment }
function SetAlpha(Color32: TColor32; NewAlpha: Integer): TColor32;

{ Color space conversions }
function HSLtoRGB(H, S, L: Single): TColor32;
procedure RGBtoHSL(RGB: TColor32; var H, S, L : Single);

{ Color algebra }
function ColorMix(C1, C2: TColor32; W1, W2: Integer): TColor32;
function ColorAdd(C1, C2: TColor32): TColor32;
function ColorSub(C1, C2: TColor32): TColor32;
function ColorModulate(C1, C2: TColor32): TColor32;
function ColorMax(C1, C2: TColor32): TColor32;
function ColorMin(C1, C2: TColor32): TColor32;

type
  { TThreadPersistent }
  { TThreadPersistent is an ancestor for TBitmap32 object. In addition to
    TPersistent class, it provides thread-safe locking and change notification }
  TThreadPersistent = class(TPersistent)
  private
    FLock: TRTLCriticalSection;
    FLockCount: Integer;
    FUpdateCount: Integer;
    FOnChange: TNotifyEvent;
  public
    constructor Create; virtual;
    destructor Destroy; override;
    procedure Changed(Sender: TObject); virtual;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Lock;
    procedure Unlock;
    property LockCount: Integer read FLockCount;
    property UpdateCount: Integer read FUpdateCount;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TFont32 }
  { TFont32 adds an Escapement property to standart TFont class }
  TFont32 = class(TFont)
  private
    FEscapement: Integer;
    procedure SetEscapement(const Value: Integer);
  published
    property Escapement: Integer read FEscapement write SetEscapement;
  end;

  { TBitmap32 }
  { This is the core of Graphics32 unit. The bitmap32 class is responsible
    for storage of bitmaps, as well as for drawing in them. }
  TBitmap32 = class(TThreadPersistent)
  private
    FAutoAlphaMult: Boolean;
    FBitmapInfo: TBitmapInfo;
    FBitmapInfoHeader: TBitmapInfoHeader;
    FBits: PColor32Array;
    FFont: TFont32;
    FHandle: THandle;
    FHeight: Integer;
    FHDC: HDC;
    FMasterAlpha: Byte;
    FOuterColor: TColor32; // the value returned when accessing outer areas
    FSmoothResize: Boolean;
    FTransparent: Boolean;
    FTransparentColor: TColor32;
    FWidth: Integer;
    FUseAlpha: Boolean;
    function  GetPixel(X, Y: Integer): TColor32;
    function  GetPixelS(X, Y: Integer): TColor32;
    function  GetPixelPtr(X, Y: Integer): PColor32;
    function  GetScanLine(Y: Integer): PColor32;
    procedure SetAutoAlphaMult(const Value: Boolean);
    procedure SetFont(Value: TFont32);
    procedure SetMasterAlpha(const Value: Byte);
    procedure SetHeight(Value: Integer);
    procedure SetPixel(X, Y: Integer; Value: TColor32);
    procedure SetPixelS(X, Y: Integer; Value: TColor32);
    procedure SetSmoothResize(Value: Boolean);
    procedure SetTransparent(const Value: Boolean);
    procedure SetTransparentColor(const Value: TColor32);
    procedure SetUseAlpha(const Value: Boolean);
    procedure SetWidth(Value: Integer);
  protected
    FontHandle: HFont;
    function ClipLine(var X0, Y0, X1, Y1: Integer): Boolean;
    function ClipLineF(var X0, Y0, X1, Y1: Single; MinX, MaxX, MinY, MaxY: Single): Boolean;
    procedure FontChanged(Sender: TObject);
    procedure SET_T(Ptr: PColor32; A, C: Integer);
    procedure SET_T256(X, Y: Integer; C: Integer);
    procedure SET_TS256(X, Y: Integer; C: Integer);
    procedure ReadData(Stream: TStream); virtual;
    procedure WriteData(Stream: TStream); virtual;
    procedure DefineProperties(Filer: TFiler); override;
  public
    constructor Create; override;
    destructor Destroy; override;

    procedure Assign(Source: TPersistent); override;
    procedure AssignTo(Dst: TPersistent); override;
    procedure SetSize(NewWidth, NewHeight: Integer); overload; virtual;
    procedure SetSize(Source: TBitmap32); overload; virtual;
    procedure Resize(NewWidth, NewHeight: Integer); virtual;
    function Empty: Boolean;
    procedure Clear; overload;
    procedure Clear(FillColor: TColor32); overload;

    procedure LoadFromStream(Stream: TStream);
    procedure SaveToStream(Stream: TStream);
    procedure LoadFromFile(const FileName: string);
    procedure SaveToFile(const FileName: string);

    procedure PreMultAlpha;
    procedure ResetAlpha;
    procedure DrawOpaque(hDst: HDC); overload;
    procedure DrawOpaque(hDst: HDC; X, Y: Integer); overload; // BitBlt to hDst
    procedure DrawOpaque(hDst: HDC; const Dst: TRect); overload;
    procedure DrawOpaque(hDst: HDC; const Dst, Src: TRect); overload;
    procedure DrawTransparent(hDst: HDC); overload;
    procedure DrawTransparent(hDst: HDC; X, Y: Integer); overload;
    procedure DrawTransparent(hDst: HDC; const Dst: TRect); overload;
    procedure DrawTransparent(hDst: HDC; const Dst, Src: TRect); overload;
    procedure DrawBlend(hDst: HDC); overload;
    procedure DrawBlend(hDst: HDC; X, Y: Integer); overload;
    procedure DrawBlend(hDst: HDC; const Dst: TRect); overload;
    procedure DrawBlend(hDst: HDC; const Dst, Src: TRect); overload;
    procedure Draw(hDst: HDC); overload;
    procedure Draw(hDst: HDC; X, Y: Integer); overload;
    procedure Draw(hDst: HDC; const Dst: TRect); overload;
    procedure Draw(hDst: HDC; const Dst, Src: TRect); overload;

    procedure SetPixelT(X, Y: Integer; Value: TColor32); overload;
    procedure SetPixelT(var Ptr: PColor32; Value: TColor32); overload;
    procedure SetPixelTS(X, Y: Integer; Value: TColor32);
    procedure SetPixelF(X, Y: Single; Value: TColor32);
    procedure SetPixelFS(X, Y: Single; Value: TColor32);

    procedure DrawHorzLine(X1, Y, X2: Integer; Value: TColor32);
    procedure DrawHorzLineS(X1, Y, X2: Integer; Value: TColor32);
    procedure DrawHorzLineT(X1, Y, X2: Integer; Value: TColor32);
    procedure DrawHorzLineTS(X1, Y, X2: Integer; Value: TColor32);

    procedure DrawVertLine(X, Y1, Y2: Integer; Value: TColor32);
    procedure DrawVertLineS(X, Y1, Y2: Integer; Value: TColor32);
    procedure DrawVertLineT(X, Y1, Y2: Integer; Value: TColor32);
    procedure DrawVertLineTS(X, Y1, Y2: Integer; Value: TColor32);

    procedure DrawLine(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineS(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineT(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineTS(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineA(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineAS(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure DrawLineF(X1, Y1, X2, Y2: Single; Value: TColor32);
    procedure DrawLineFS(X1, Y1, X2, Y2: Single; Value: TColor32);

    procedure FillRect(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure FillRectS(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure FillRectT(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure FillRectTS(X1, Y1, X2, Y2: Integer; Value: TColor32);

    procedure FrameRectS(X1, Y1, X2, Y2: Integer; Value: TColor32);
    procedure FrameRectTS(X1, Y1, X2, Y2: Integer; Value: TColor32);

    procedure RaiseRectTS(X1, Y1, X2, Y2: Integer; Contrast: Integer);

    procedure UpdateFont;
    procedure TextOut(X, Y: Integer; const Text: string); overload;
    procedure TextOut(X, Y: Integer; const ClipRect: TRect; const Text: string); overload;
    procedure TextOut(ClipRect: TRect; const Flags: Cardinal; const Text: string); overload;
    function TextExtent(const Text: string): TSize;
    function TextHeight(const Text: string): Integer;
    function TextWidth(const Text: string): Integer;

    procedure ColorToGrayscale(ColorSrc: TBitmap32 = nil);
    procedure AlphaToGrayscale(AlphaSrc: TBitmap32 = nil);
    procedure IntensityToAlpha(IntensitySrc: TBitmap32 = nil);
    procedure Invert(Src: TBitmap32 = nil);

    property Handle: HDC read FHDC;
    property Bits: PColor32Array read FBits; // use this array for direct access

    property ScanLine[Y: Integer]: PColor32 read GetScanLine;
    property Pixel[X, Y: Integer]: TColor32 read GetPixel write SetPixel; default;
    property PixelS[X, Y: Integer]: TColor32 read GetPixelS write SetPixelS;
    property PixelPtr[X, Y: Integer]: PColor32 read GetPixelPtr;
    property Height: Integer read FHeight write SetHeight;
    property Width: Integer read FWidth write SetWidth;
  published
    property AutoAlphaMult: Boolean read FAutoAlphaMult write SetAutoAlphaMult;
    property Font: TFont32 read FFont write SetFont;
    property MasterAlpha: Byte read FMasterAlpha write SetMasterAlpha;
    property OuterColor: TColor32 read FOuterColor write FOuterColor default $00000000;
    property SmoothResize: Boolean read FSmoothResize write SetSmoothResize;
    property Transparent: Boolean read FTransparent write SetTransparent;
    property TransparentColor: TColor32 read FTransparentColor write SetTransparentColor;
    property UseAlpha: Boolean read FUseAlpha write SetUseAlpha;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { TSprite }
  { A bitmap container used as a sprite by the TImage32 component }
  TSprite = class(TCollectionItem)
  private
    FBitmap: TBitmap32;
    FPositionX: Integer;
    FPositionY: Integer;
    FVisible: Boolean;
    procedure SetBitmap(Value: TBitmap32);
    procedure SetPosition(const Value: TPoint);
    procedure SetPositionX(Value: Integer);
    procedure SetPositionY(Value: Integer);
    procedure SetVisible(Value: Boolean);
    function GetPosition: TPoint;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Changed(Sender: TObject);
    property Position: TPoint read GetPosition write SetPosition;
  published
    property Bitmap: TBitmap32 read FBitmap write SetBitmap;
    property PositionX: Integer read FPositionX write SetPositionX;
    property PositionY: Integer read FPositionY write SetPositionY;
    property Visible: Boolean read FVisible write SetVisible;
  end;

  TImage32 = class;

  { TSpriteSet }
  { A collection of TSprite objects }
  TSpriteSet = class(TCollection)
  private
    FImage: TImage32;
    FOnChange: TNotifyEvent;
    function GetItem(Index: Integer): TSprite;
    procedure SetItem(Index: Integer; Value: TSprite);
  protected
    procedure Changed(Sender: TObject);
    function GetOwner: TPersistent; override;
    procedure SpriteChanged(Sprite: TSprite);
  public
    constructor Create(Image: TImage32);
    function Add: TSprite;
    property Image: TImage32 read FImage;
    property Items[Index: Integer]: TSprite read GetItem write SetItem; default;
    property OnChange: TNotifyEvent read FOnChange write FOnChange;
  end;

  { Color manager entries }
  PColorEntry = ^TColorEntry;
  TColorEntry = record
    Name: string[31];
    Color: TColor32;
  end;

  { TImage32 }
  { An image displaying VCL component for TBitmap32 object }
  TImage32 = class(TCustomControl)
  private
    FBitmap: TBitmap32;
    FBitmapAlign: TBitmapAlign;
    FScaleMode: TScaleMode;
    FScale: Single;
    FSprites: TSpriteSet;
    FSpritesOrigin: TSpriteOrigin;
    procedure BitmapChanged(Sender: TObject);
    procedure SpritesChanged(Sender: TObject);
    procedure SetBitmap(Value: TBitmap32);
    procedure SetScale(Value: Single);
    procedure SetBitmapAlign(Value: TBitmapAlign);
    procedure SetScaleMode(Value: TScaleMode);
    procedure SetSprites(Value: TSpriteSet);
    procedure SetSpritesOrigin(Value: TSpriteOrigin);
    procedure WMWindowPosChanged(var Message: TWMWindowPosChanged); message WM_WINDOWPOSCHANGED;
  protected
    UpdateCount: Integer;
    procedure Changed(Sender: TObject); virtual;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    procedure BeginUpdate;
    procedure EndUpdate;
    procedure Paint; override;
    procedure Resize; override;
    procedure SetupBitmap(DoClear: Boolean = False; ClearColor: TColor32 = $FF000000); virtual;
  published
    property Align;
    property Anchors;
    property Color;
    property Constraints;
    property Cursor;
    property Bitmap: TBitmap32 read FBitmap write SetBitmap;
    property BitmapAlign: TBitmapAlign read FBitmapAlign write SetBitmapAlign;
    property DoubleBuffered;
    property ParentColor;
    property ParentShowHint;
    property PopupMenu;
    property Scale: Single read FScale write SetScale;
    property ScaleMode: TScaleMode read FScaleMode write SetScaleMode;
    property ShowHint;
    property Sprites: TSpriteSet read FSprites write SetSprites;
    property SpritesOrigin: TSpriteOrigin read FSpritesOrigin write SetSpritesOrigin;
    property Visible;
    property OnClick;
    property OnContextPopup;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnMouseDown;
    property OnMouseMove;
    property OnMouseUp;
  end;

  { TColorManager }
  { TColorManager provides support for TColor32 in design time }
  TColorManager = class(TList)
  public
    destructor Destroy; override;
    procedure AddColor(AName: string; AColor: TColor32);
    procedure EnumColors(Proc: TGetStrProc);
    function FindColor(AName: string): TColor32;
    function GetColor(AName: string): TColor32;
    function GetColorName(AColor: TColor32): string;
    procedure RegisterDefaultColors;
    procedure RemoveColor(AName: string);
  end;

var
  ColorManager: TColorManager;

procedure RegisterColor(AName: string; AColor: TColor32);
procedure UnregisterColor(AName: string);


{ Gamma support for antialiasing }
var
  GAMMA_TABLE: array [Byte] of Byte;
procedure SetGamma(Gamma: Single = 0.7);


implementation


uses Math, TypInfo, Clipbrd;

const
  { Component masks }
  _R   : TColor32 = $00FF0000;
  _G   : TColor32 = $0000FF00;
  _B   : TColor32 = $000000FF;
  _RGB : TColor32 = $00FFFFFF;

var
  Bitmap32CounterLock: TRTLCriticalSection;
  StockFont: HFONT;

{ Clamp function restricts Value to [0..255] range }
function Clamp(Value: Integer): TColor32;
begin
  if Value < 0 then Result := 0
  else if Value > 255 then Result := 255
  else Result := Value;
end;

{ An analogue of FillChar for 32 bit values }
procedure FillLongword(var X; Count: Integer; Value: Longword);
asm
// EAX = X
// EDX = Count
// ECX = Value
        PUSH    EDI

        MOV     EDI,EAX  // Point EDI to destination              
        MOV     EAX,ECX
        MOV     ECX,EDX
        TEST    ECX,ECX
        JS      @@exit

        REP     STOSD    // Fill count dwords
@@exit:
        POP     EDI
end;

{ Exchange two 32-bit values }
procedure Swap(var A, B: Integer);
asm
// EAX = [A]
// EDX = [B]
        MOV     ECX,[EAX]     // ECX := [A]
        XCHG    ECX,[EDX]     // ECX <> [B];
        MOV     [EAX],ECX     // [A] := ECX
end;

{ Exhange A <-> B only if B < A }
procedure TestSwap(var A, B: Integer);
asm
// EAX = [A]
// EDX = [B]
        MOV     ECX,[EAX]     // ECX := [A]
        CMP     ECX,[EDX]     // ECX <= [B]? Exit
        JLE     @@exit
        XCHG    ECX,[EDX]     // ECX <-> [B];
        MOV     [EAX],ECX     // [A] := ECX
@@exit:
end;

{ Exhange A <-> B only if B < A then restrict both to [0..Size-1] range }
{ returns true if resulting range has common points with [0..Size-1] range }
function TestClip(var A, B: Integer; Size: Integer): Boolean;
begin
  TestSwap(A, B); // now A = min(A,B) and B = max(A, B)
  if A < 0 then A := 0;
  if B >= Size then B := Size - 1;
  Result := B >= A;
end;

type
  TGraphicAccess = class(TGraphic);


{ Color construction and conversion functions }

{ Colorswap exchanges ARGB <-> ABGR and fill A with $FF }
function ColorSwap(WinColor: TColor): TColor32; register; assembler;
asm
// EAX = WinColor
        MOV     ECX,EAX         // this function swaps R and B bytes in ABGR
        SHR     EAX,16
        XCHG    AL,CL
        MOV     AH,$FF          // and writes $FF into A component
        SHL     EAX,16
        MOV     AX, CX
end;

function Color32(WinColor: TColor): TColor32; overload;
begin
  WinColor := ColorToRGB(WinColor);
  Result := ColorSwap(WinColor);
end;

function Color32(R, G, B: Byte; A: Byte = $FF): TColor32; overload;
begin
  Result := A shl 24 + R shl 16 + G shl 8 + B;
end;

function Color32(Index: Byte; var Palette: TPalette32): TColor32; overload;
begin
  Result := Palette[Index];
end;

function Gray32(Intensity: Byte; Alpha: Byte = $FF): TColor32;
begin
  Result := TColor32(Alpha) shl 24 + TColor32(Intensity) shl 16 +
    TColor32(Intensity) shl 8 + TColor32(Intensity);
end;

function WinColor(Color32: TColor32): TColor;
begin
  Result := (Color32 and _R shr 16) or (Color32 and _G) or
  (Color32 and _B shl 16);
end;

{ Component access }

function RedComponent(Color32: TColor32): Integer;
begin
  Result := Color32 and _R shr 16;
end;

function GreenComponent(Color32: TColor32): Integer;
begin
  Result := Color32 and _G shr 8;
end;

function BlueComponent(Color32: TColor32): Integer;
begin
  Result := Color32 and _B;
end;

function AlphaComponent(Color32: TColor32): Integer;
begin
  Result := Color32 shr 24;
end;

function Intensity(Color32: TColor32): Integer;
begin
  Result := ((Color32 and _R shr 16) * 54 + (Color32 and _G shr 8) * 182 +
    (Color32 and _B) * 19) shr 8;
end;

{ Alpha adjustment }

function SetAlpha(Color32: TColor32; NewAlpha: Integer): TColor32;
begin
  Result := (Color32 and _RGB) or (Clamp(NewAlpha) shl 24);
end;

function MultAlpha(Color32: TColor): TColor32;
var
  r, g, b, a: TColor32;
begin
  a := Color32 shr 24;
  r := (Color32 and _R) * a div $FF and _R;
  g := (Color32 and _G) * a div $FF and _G;
  b := (Color32 and _B) * a div $FF and _B;
  Result := r + g + b + a shl 24;
end;

{ Color space conversions }

function HSLtoRGB(H, S, L: Single): TColor32;
var
  M1, M2: Single;
  R, G, B: Byte;

  function HueToColourValue(Hue: Single): Byte;
  var
    V: Double;
  begin
    Hue := Hue - Floor(Hue);

    if 6 * Hue < 1 then V := M1 + (M2 - M1) * Hue * 6
    else if 2 * Hue < 1 then V := M2
    else if 3 * Hue < 2 then V := M1 + (M2 - M1) * (2 / 3 - Hue) * 6
    else V := M1;
    Result := Round(255 * V);
  end;
begin
  if S = 0 then
  begin
    R := Round(255 * L);
    G := R;
    B := R;
  end
  else
  begin
    if L <= 0.5 then M2 := L * (1 + S)
    else M2 := L + S - L * S;
    M1 := 2 * L - M2;
    R := HueToColourValue(H - 1 / 3);
    G := HueToColourValue(H);
    B := HueToColourValue(H + 1 / 3)
  end;
  Result := Color32(R, G, B, 255);
end;

procedure RGBtoHSL(RGB: TColor32; var H, S, L : Single);
var
  R, G, B, D, Cmax, Cmin: Single;
begin
  R := RedComponent(RGB) / 255;
  G := GreenComponent(RGB) / 255;
  B := BlueComponent(RGB) / 255;
  Cmax := Max (R, Max (G, B));
  Cmin := Min (R, Min (G, B));
  L := (Cmax + Cmin) / 2;

  if Cmax = Cmin then
  begin
    H := 0;
    S := 0
  end
  else
  begin
    D := Cmax - Cmin;
    if L < 0.5 then S := D / (Cmax + Cmin)
    else S := D / (2 - Cmax - Cmin);
    if R = Cmax then H := (G - B) / D
    else
      if G = Cmax then H  := 2 + (B - R) /D
      else H := 4 + (R - G) / D;
    H := H / 6;
    if H < 0 then H := H + 1
  end
end;

{ Color algebra }

function ColorMix(C1, C2: TColor32; W1, W2: Integer): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := (a1 * W1 + a2 * W2) div 100;
  r1 := (r1 * W1 + r2 * W2) div 100;
  g1 := (g1 * W1 + g2 * W2) div 100;
  b1 := (b1 * W1 + b2 * W2) div 100;
  Result := Color32(Clamp(r1), Clamp(g1), Clamp(b1), Clamp(a1));
end;

function ColorAdd(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := a1 + a2;
  r1 := r1 + r2;
  g1 := g1 + g2;
  b1 := b1 + b2;
  Result := Color32(Clamp(r1), Clamp(g1), Clamp(b1), Clamp(a1));
end;

function ColorSub(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := a1 - a2;
  r1 := r1 - r2;
  g1 := g1 - g2;
  b1 := b1 - b2;
  Result := Color32(Clamp(r1), Clamp(g1), Clamp(b1), Clamp(a1));
end;

function ColorModulate(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := a1 * a2 shr 8;
  r1 := r1 * r2 shr 8;
  g1 := g1 * g2 shr 8;
  b1 := b1 * b2 shr 8;
  Result := Color32(r1, g1, b1, a1);
end;

function ColorMax(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := Max(a1, a2);
  r1 := Max(r1, r2);
  g1 := Max(g1, g2);
  b1 := Max(b1, b2);
  Result := Color32(r1, g1, b1, a1);
end;

function ColorMin(C1, C2: TColor32): TColor32;
var
  r1, g1, b1, a1: Integer;
  r2, g2, b2, a2: Integer;
begin
  a1 := C1 shr 24;
  r1 := C1 and _R shr 16;
  g1 := C1 and _G shr 8;
  b1 := C1 and _B;
  a2 := C2 shr 24;
  r2 := C2 and _R shr 16;
  g2 := C2 and _G shr 8;
  b2 := C2 and _B;
  a1 := Max(a1, a2);
  r1 := Max(r1, r2);
  g1 := Max(g1, g2);
  b1 := Max(b1, b2);
  Result := Color32(r1, g1, b1, a1);
end;

{ TThreadPersistent }

procedure TThreadPersistent.BeginUpdate;
begin
  Inc(FUpdateCount);
end;

procedure TThreadPersistent.Changed(Sender: TObject);
begin
  if (FUpdateCount = 0) and Assigned(FOnChange) then FOnChange(Sender);
end;

constructor TThreadPersistent.Create;
begin
  InitializeCriticalSection(FLock);
end;

destructor TThreadPersistent.Destroy;
begin
  DeleteCriticalSection(FLock);
  inherited;
end;

procedure TThreadPersistent.EndUpdate;
begin
  Assert(FUpdateCount > 0, 'Error in TThreadPersistent.EndUpdate');
  Dec(FUpdateCount);
  if FUpdateCount = 0 then Changed(Self);
end;

procedure TThreadPersistent.Lock;
begin
  EnterCriticalSection(Bitmap32CounterLock);
  Inc(FLockCount);
  LeaveCriticalSection(Bitmap32CounterLock);
  EnterCriticalSection(FLock);
end;

procedure TThreadPersistent.Unlock;
begin
  LeaveCriticalSection(FLock);
  EnterCriticalSection(Bitmap32CounterLock);
  Dec(FLockCount);
  LeaveCriticalSection(Bitmap32CounterLock);
end;


{ TFont32 }

procedure TFont32.SetEscapement(const Value: Integer);
begin
  FEscapement := Value;
  Changed;
end;


{ TBitmap32 }

constructor TBitmap32.Create;
begin
  inherited;
  FOuterColor := $00000000;  // by default as full transparency black
  FFont := TFont32.Create;
  FFont.OnChange := FontChanged;
  FFont.OwnerCriticalSection := @FLock;
  FMasterAlpha := $FF;
end;

destructor TBitmap32.Destroy;
begin
  Lock;
  try
    FFont.Free;
    SetSize(0, 0);
  finally
    Unlock;
  end;
  inherited;
end;

procedure TBitmap32.SetSize(NewWidth, NewHeight: Integer);
begin
  if (NewWidth <= 0) then NewWidth := 0;
  if (NewHeight <= 0) then NewHeight := 0;
  if (NewWidth = Width) and (NewHeight = Height) then Exit;

  BeginUpdate;
  try
    // destroy the old bitmap if necessary
    if not Empty then
    begin
      DeleteDC(FHDC);
      DeleteObject(FHandle);
      if FontHandle > 0 then
      begin
        DeleteObject(FontHandle);
        FontHandle := 0;
      end;
      FHDC := 0;
      FHandle := 0;
      FWidth := 0;
      FHeight := 0;
    end;

    // allocate a new DIB
    if (NewWidth > 0) and (NewHeight > 0) then
    begin
      FWidth := 0;
      FHeight := 0;
      with FBitmapInfoHeader do
      begin
        biSize := SizeOf(TBitmapInfoHeader);
        biWidth := NewWidth;
        biHeight := -NewHeight; // minus means that the top row number is zero
        biPlanes := 1;
        biBitCount := 32;
        biCompression := BI_RGB;
      end;
      FBitmapInfo.bmiHeader := FBitmapInfoHeader;

      FHandle := CreateDIBSection(0, FBitmapInfo, DIB_RGB_COLORS,
        Pointer(FBits), 0, 0);
      if FBits = nil then
        raise Exception.Create('Can''t allocate DIB handle');

      FHDC := CreateCompatibleDC(0);
      if FHDC = 0 then
      begin
        DeleteObject(FHandle);
        FHandle := 0;
        raise Exception.Create('Can''t create compatible DC');
      end;

      if SelectObject(FHDC, FHandle) = 0 then
      begin
        DeleteDC(FHDC);
        DeleteObject(FHandle);
        FHDC := 0;
        FHandle := 0;
        raise Exception.Create('Can''t select an object into DC');
      end;
    end;
    FWidth := NewWidth;
    FHeight := NewHeight;
  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.SetSize(Source: TBitmap32);
begin
  if Source <> nil then SetSize(Source.Width, Source.Height)
  else SetSize(0, 0);
end;

procedure TBitmap32.Resize(NewWidth, NewHeight: Integer);
var
  Bmp: TBitmap32;
begin
  if (Width <> NewWidth) or (Height <> NewHeight) then
  begin
    BeginUpdate;
    Bmp := TBitmap32.Create;
    try
      Bmp.Assign(Self);
      SetSize(NewWidth, NewHeight);
      Bmp.DrawOpaque(Handle, Rect(0, 0, Width, Height));
    finally
      Bmp.Free;
      EndUpdate;
    end;
  end;
end;

function TBitmap32.Empty: Boolean;
begin
  Result := (FHandle = 0);
end;

procedure TBitmap32.Clear;
begin
  Clear(clBlack32);
end;

procedure TBitmap32.Clear(FillColor: TColor32);
begin
  if Empty then Exit;
  BeginUpdate;
  try
    FillLongword(Bits[0], Width * Height, FillColor);
  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.Assign(Source: TPersistent);
var
  Canvas: TCanvas;
  Picture: TPicture;

  procedure AssignFromBitmap(SrcBmp: TBitmap);
  begin
    SetSize(SrcBmp.Width, SrcBmp.Height);
    if Empty then Exit;
    BitBlt(Handle, 0, 0, Width, Height, SrcBmp.Canvas.Handle, 0, 0, SRCCOPY);
  end;
begin
  BeginUpdate;
  try
    if Source = nil then
    begin
      SetSize(0, 0);
      Exit;
    end;

    if Source is TBitmap32 then
    begin
      SetSize(TBitmap32(Source));
      Move(TBitmap32(Source).Bits[0], Bits[0], Width * Height * 4);
      Exit;
    end;

    if Source is TBitmap then
    begin
      AssignFromBitmap(TBitmap(Source));
      Exit;
    end;

    if Source is TPicture then
    begin
      with TPicture(Source) do
      begin
        if TPicture(Source).Graphic is TBitmap then
          AssignFromBitmap(TBitmap(TPicture(Source).Graphic))
        else
        begin
          SetSize(Width, Height);
          if Empty then Exit;
          Canvas := TCanvas.Create;
          try
            Canvas.Handle := Self.Handle;
            TGraphicAccess(Graphic).Draw(Canvas, Rect(0, 0, Width, Height));
          finally
            Canvas.Free;
          end;
        end;
      end;
      Exit;
    end;

    if Source is TClipboard then
    begin
      Picture := TPicture.Create;
      try
        Picture.Assign(TClipboard(Source));
        SetSize(Picture.Width, Picture.Height);
        if Empty then Exit;
        Canvas := TCanvas.Create;
        try
          Canvas.Handle := Self.Handle;
          TGraphicAccess(Picture.Graphic).Draw(Canvas, Rect(0, 0, Width, Height));
        finally
          Canvas.Free;
        end;
      finally
        Picture.Free;
      end;
      Exit;
    end;

    inherited; // default handler

  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.AssignTo(Dst: TPersistent);
var
  Bmp: TBitmap;
begin
  if Dst is TPicture then
  begin
    Bmp := TPicture(Dst).Bitmap;
    Bmp.HandleType := bmDIB;
    Bmp.PixelFormat := pf32Bit;
    Bmp.Width := Width;
    Bmp.Height := Height;
    DrawOpaque(Bmp.Canvas.Handle);
    Exit;
  end;

  if Dst is TBitmap then
  begin
    Bmp := TBitmap(Dst);
    Bmp.HandleType := bmDIB;
    Bmp.PixelFormat := pf32Bit;
    Bmp.Width := Width;
    Bmp.Height := Height;
    DrawOpaque(Bmp.Canvas.Handle);
    Exit;
  end;

  if Dst is TClipboard then
  begin
    Bmp := TBitmap.Create;
    try
      Bmp.HandleType := bmDIB;
      Bmp.PixelFormat := pf32Bit;
      Bmp.Width := Width;
      Bmp.Height := Height;
      DrawOpaque(Bmp.Canvas.Handle);
      TClipboard(Dst).Assign(Bmp);
    finally
      Bmp.Free;
    end;
    Exit;
  end;

  inherited;
end;

procedure TBitmap32.SetHeight(Value: Integer);
begin
  SetSize(Width, Value);
end;

procedure TBitmap32.SetWidth(Value: Integer);
begin
  SetSize(Value, Height);
end;

procedure TBitmap32.SetPixel(X, Y: Integer; Value: TColor32);
begin
  Bits[X + Y * Width] := Value;
end;

procedure TBitmap32.SetPixelS(X, Y: Integer; Value: TColor32);
begin
  if (X >= 0) and (X < Width) and (Y >= 0) and (Y < Height) then
    Bits[X + Y * Width] := Value;
end;

function TBitmap32.GetScanLine(Y: Integer): PColor32;
begin
  Result := @Bits[Y * FWidth];
end;

function TBitmap32.GetPixel(X, Y: Integer): TColor32;
begin
  Result := Bits[X + Y * Width];
end;

function TBitmap32.GetPixelS(X, Y: Integer): TColor32;
begin
  if (X >= 0) and (X < Width) and (Y >= 0) and (Y < Width) then
    Result := Bits[X + Y * FWidth]
  else
    Result := OuterColor;
end;

function TBitmap32.GetPixelPtr(X, Y: Integer): PColor32;
begin
  Result := @Bits[X + Y * FWidth];
end;

procedure TBitmap32.DrawOpaque(hDst: HDC);
begin
  DrawOpaque(hDst, 0, 0);
end;

procedure TBitmap32.DrawOpaque(hDst: HDC; X, Y: Integer);
begin
  if Empty then Exit;
  if (x >= 0) and (y >= 0) then
    BitBlt(hDst, X, Y, Width, Height, FHDC, 0, 0, SRCCOPY)
  else
    DrawOpaque(hDst, Rect(X, Y, X + Width, Y + Height));
end;

procedure TBitmap32.DrawOpaque(hDst: HDC; const Dst: TRect);
begin
  if Empty then Exit;
  if (Width = Dst.Right - Dst.Left + 1) and
    (Height = Dst.Bottom - Dst.Top + 1) and
    (Dst.Left >= 0) and
    (Dst.Right >= 0) then
    DrawOpaque(hDst, Dst.Left, Dst.Top)
  else
    DrawOpaque(hDst, Dst, Rect(0, 0, Width, Height));
end;

procedure TBitmap32.DrawOpaque(hDst: HDC; const Dst, Src: TRect);
var
  Pt: TPoint;
begin
  if Empty then Exit;
  if FSmoothResize then
  begin
    GetBrushOrgEx(hDst, pt);
    SetStretchBltMode(hDst, STRETCH_HALFTONE);
    SetBrushOrgEx(hDst, pt.x, pt.y, @pt);
  end    
  else
    SetStretchBltMode(hDst, STRETCH_DELETESCANS);

 StretchDIBits(hDst, Dst.Left, Dst.Top, Dst.Right - Dst.Left,
   Dst.Bottom - Dst.Top, Src.Left, Src.Top, Src.Right - Src.Left,
   Src.Bottom - Src.Top, Bits, FBitmapInfo, DIB_RGB_COLORS, SRCCOPY);
end;

procedure TBitmap32.DrawTransparent(hDst: HDC);
begin
  DrawTransparent(hDst, 0, 0);
end;

procedure TBitmap32.DrawTransparent(hDst: HDC; X, Y: Integer);
begin
  if Empty then Exit;
  TransparentBlt(hDst, X, Y, FWidth + X, FHeight + X,
    Handle, 0, 0, FWidth, FHeight,
    WinColor(TransparentColor));
end;

procedure TBitmap32.DrawTransparent(hDst: HDC; const Dst: TRect);
begin
  if Empty then Exit;
  TransparentBlt(
    hDst, Dst.Left, Dst.Top, Dst.Right - Dst.Left,
    Dst.Bottom - Dst.Top,
    Handle, 0, 0, FWidth, FHeight,
    WinColor(TransparentColor));
end;

procedure TBitmap32.DrawTransparent(hDst: HDC; const Dst, Src: TRect);
begin
  if Empty then Exit;
  TransparentBlt(
    hDst, Dst.Left, Dst.Top, Dst.Right - Dst.Left, Dst.Bottom - Dst.Top,
    Handle, Src.Left, Src.Top, Src.Right - Src.Left, Src.Bottom - Src.Top,
    WinColor(TransparentColor));
end;

procedure TBitmap32.DrawBlend(hDst: HDC);
begin
  DrawBlend(hDst, 0, 0);
end;

procedure TBitmap32.DrawBlend(hDst: HDC; X, Y: Integer);
begin
  DrawBlend(hDst, Rect(X, Y, X + Width, Y + Height), Rect(0, 0, Width, Height))
end;

procedure TBitmap32.DrawBlend(hDst: HDC; const Dst: TRect);
begin
  DrawBlend(hDst, Dst, Rect(0, 0, Width, Height))
end;

procedure TBitmap32.DrawBlend(hDst: HDC; const Dst, Src: TRect);
var
  BlendFunc: TBlendFunction;
  BMP: TBitmap32;
begin
  if Empty then Exit;

  BlendFunc.BlendOp := 0;
  BlendFunc.BlendFlags := 0;
  BlendFunc.SourceConstantAlpha := MasterAlpha;
  BlendFunc.AlphaFormat := 1;

  if not AutoAlphaMult then
    AlphaBlend(
      hDst, Dst.Left, Dst.Top, Dst.Right - Dst.Left, Dst.Bottom - Dst.Top,
      Handle, Src.Left, Src.Top, Src.Right - Src.Left, Src.Bottom - Src.Top,
      BlendFunc)
  else
  begin
    BMP := TBitmap32.Create;
    try
      BMP.Assign(Self);
      BMP.PreMultAlpha;
      AlphaBlend(
        hDst, Dst.Left, Dst.Top, Dst.Right - Dst.Left, Dst.Bottom - Dst.Top,
        BMP.Handle, Src.Left, Src.Top, Src.Right - Src.Left, Src.Bottom - Src.Top,
        BlendFunc);
    finally
      BMP.Free;
    end;
  end;
end;

procedure TBitmap32.Draw(hDst: HDC);
begin
  Draw(hDst, 0, 0);
end;

procedure TBitmap32.Draw(hDst: HDC; X, Y: Integer);
begin
  if Empty then exit;
  if Transparent then DrawTransparent(hDst, X, Y)
  else if UseAlpha then DrawBlend(hDst, X, Y)
  else DrawOpaque(hDst, X, Y);
end;

procedure TBitmap32.Draw(hDst: HDC; const Dst: TRect);
begin
  if Empty then exit;
  if Transparent then DrawTransparent(hDst, Dst)
  else if UseAlpha then DrawBlend(hDst, Dst)
  else DrawOpaque(hDst, Dst);
end;

procedure TBitmap32.Draw(hDst: HDC; const Dst, Src: TRect);
begin
  if Empty then exit;
  if Transparent then DrawTransparent(hDst, Dst, Src)
  else if UseAlpha then DrawBlend(hDst, Dst, Src)
  else DrawOpaque(hDst, Dst, Src);
end;

procedure TBitmap32.PreMultAlpha;
var
  I: Integer;
  P: PColor32;
begin
  P := Pointer(FBits);
  for I := 0 to Width * Height - 1 do
  begin
    P^ := MultAlpha(P^);
    Inc(P);
  end;
end;

procedure TBitmap32.ResetAlpha;
var
  I: Integer;
  P: PColor32;
begin
  P := Pointer(FBits);
  for I := 0 to Width * Height - 1 do
  begin
    P^ := P^ or $FF000000;
    Inc(P)
  end;
end;

function Combine(CF, CB: TColor32; A: TColor32): TColor32;
var
  X: TColor32;
begin
  X := A xor 255;
  Result := ((CB and $FF0000) * X + (CF and $FF0000) * A) and $FF000000;
  Result := Result + ((CB and $FF00) * X + (CF and $FF00) * A) and $FFFF0000;
  Result := Result + ((CB and $FF) * X + (CF and $FF) * A);
  Result := Result shr 8 or $FF000000;
end;

procedure TBitmap32.SET_T(Ptr: PColor32; A, C: Integer);
var
  B: TColor32;
begin
  B := Ptr^; // background color
  Ptr^ := Combine(C, B, A);
end;

procedure TBitmap32.SetPixelT(X, Y: Integer; Value: TColor32);
var
  a: Integer;
begin
  a := Value shr 24; // opacity
  if a = 0 then Exit;
  if a = $FF then
  begin
    Bits[X + Y * Width] := Value or $FF000000;
    Exit;
  end;
  SET_T(@Bits[X + Y * Width], a, Value);
end;

procedure TBitmap32.SetPixelT(var Ptr: PColor32; Value: TColor32);
var
  a: Integer;
begin
  a := Value shr 24; // opacity
  if a > 0 then
  begin
    if a = $FF then Ptr^ := Value or $FF000000
    else SET_T(Ptr, a, Value);
  end;
  Inc(Ptr);
end;

procedure TBitmap32.SetPixelTS(X, Y: Integer; Value: TColor32);
begin
  if (X >= 0) and (X < Width) and (Y >= 0) and (Y < Width) then
    SetPixelT(X, Y, Value);
end;

procedure TBitmap32.SET_T256(X, Y, C: Integer);
var
  flrx, flry, celx, cely: Longword;
  P: PColor32;
  A: TColor32;
begin
  A := C shr 24;  // opacity

  flrx := X and $FF;
  flry := Y and $FF;
  if X > 0 then X := X shr 8
  else X := 0;
  if Y > 0 then Y := Y shr 8
  else Y := 0;

  celx := GAMMA_TABLE[flrx xor 255];
  cely := GAMMA_TABLE[flry xor 255];
  flrx := GAMMA_TABLE[flrx];
  flry := GAMMA_TABLE[flry];

  P := PColor32(Integer(FBits) + (X + Y * FWidth) shl 2);

  SET_T(P, A * celx * cely shr 16, C); // this is not cery accurate
  Inc(P);                              // actually instead of 'shr 16'
  SET_T(P, A * flrx * cely shr 16, C); // it should be 'div 65280'
  Inc(P, FWidth);
  SET_T(P, A * flrx * flry shr 16, C);
  Dec(P);
  SET_T(P, A * celx * flry shr 16, C);
end;

procedure TBitmap32.SET_TS256(X, Y, C: Integer);
var
  flrx, flry, celx, cely: Longword;
  P: PColor32;
  A: TColor32;
begin
  if (X < 0) or (Y < 0) then Exit;

  flrx := X and $FF;
  flry := Y and $FF;
  X := X shr 8;
  Y := Y shr 8;

  if (X < 0) or (X >= FWidth) or (Y < 0) or (Y >= FHeight) then Exit;

  A := C shr 24;  // opacity

  celx := GAMMA_TABLE[flrx xor 255];
  cely := GAMMA_TABLE[flry xor 255];
  flrx := GAMMA_TABLE[flrx];
  flry := GAMMA_TABLE[flry];

  P := PColor32(Integer(FBits) + (X + Y * FWidth) shl 2);

  SET_T(P, A * celx * cely shr 16, C);
  Inc(P);
  if X < FWidth - 1 then
    SET_T(P, A * flrx * cely shr 16, C);
  Inc(P, FWidth);
  if (X < FWidth - 1) and (Y < FHeight - 1) then
    SET_T(P, A * flrx * flry shr 16, C);
  Dec(P);
  if Y < FHeight - 1 then
    SET_T(P, A * celx * flry shr 16, C);
end;

procedure TBitmap32.SetPixelF(X, Y: Single; Value: TColor32);
begin
  SET_T256(Round(X * 256), Round(Y * 256), Value);
end;

procedure TBitmap32.SetPixelFS(X, Y: Single; Value: TColor32);
begin
  SET_TS256(Round(X * 256), Round(Y * 256), Value);
end;

procedure TBitmap32.DrawHorzLine(X1, Y, X2: Integer; Value: TColor32);
var
  P: PColor32Array;
begin
  P := Pointer(GetScanLine(Y));
  FillLongword(P[X1], X2 - X1 + 1, Value)
end;

procedure TBitmap32.DrawHorzLineS(X1, Y, X2: Integer; Value: TColor32);
begin
  if (Y >= 0) and (Y < Height) and TestClip(X1, X2, Width) then
    DrawHorzLine(X1, Y, X2, Value);
end;

procedure TBitmap32.DrawHorzLineT(X1, Y, X2: Integer; Value: TColor32);
var
  i: Integer;
  P: PColor32;
begin
  if X2 < X1 then Exit;
  P := PixelPtr[X1, Y];
  for i := X1 to X2 do SetPixelT(P, Value);
end;

procedure TBitmap32.DrawHorzLineTS(X1, Y, X2: Integer; Value: TColor32);
begin
  if (Y >= 0) and (Y < Height) and TestClip(X1, X2, Width) then
    DrawHorzLineT(X1, Y, X2, Value);
end;

procedure TBitmap32.DrawVertLine(X, Y1, Y2: Integer; Value: TColor32);
var
  i: Integer;
  P: PColor32;
begin
  if Y2 < Y1 then Exit;
  P := PixelPtr[X, Y1];
  for i := 0 to Y2 - Y1 do
  begin
    P^ := Value;
    Inc(P, Width);
  end;
end;

procedure TBitmap32.DrawVertLineS(X, Y1, Y2: Integer; Value: TColor32);
begin
  if (X >= 0) and (X < Width) and TestClip(Y1, Y2, Height) then
    DrawVertLine(X, Y1, Y2, Value);
end;

procedure TBitmap32.DrawVertLineT(X, Y1, Y2: Integer; Value: TColor32);
var
  i: Integer;
  P: PColor32;
begin
  P := PixelPtr[X, Y1];
  for i := Y1 to Y2 do
  begin
    SetPixelT(P, Value);
    Inc(P, Width - 1);
  end;
end;

procedure TBitmap32.DrawVertLineTS(X, Y1, Y2: Integer; Value: TColor32);
begin
  if (X >= 0) and (X < Width) and TestClip(Y1, Y2, Height) then
    DrawVertLineT(X, Y1, Y2, Value);
end;

procedure TBitmap32.DrawLine(X1, Y1, X2, Y2: Integer; Value: TColor32);
var
  lp1: Integer;
  x: Integer;
  dy, dx, Step, Delta: Integer;
  P: PColor32;
begin
  if Y1 > Y2 then
  begin
    Swap(X1, X2);
    Swap(Y1, Y2);
  end;

  dx := X2 - X1;
  dy := Y2 - Y1;

  if dy = 0 then
  begin
    TestSwap(X1, X2);
    DrawHorzLine(X1, Y1, X2, Value);
    Exit;
  end;

  if dx = 0 then
  begin
    DrawVertLine(X1, Y1, Y2, Value);
    Exit;
  end;

  if Abs(dx) > dy then
  begin
    Delta := dy shr 1;
    if dx >= 0 then
    begin
      P := PixelPtr[X1, Y1];
      for lp1 := X1 to X2 do
      begin
        P^ := Value;
        Inc(P);
        Delta := Delta + dy;
        if Delta > dx then
        begin
          Inc(P, Width);
          Delta := Delta - dx;
        end;
      end;
    end
    else // dx < 0
    begin
      dx := -dx;
      dy := -dy;
      P := PixelPtr[X2, Y2];
      for lp1 := X2 to X1 do
      begin
        P^ := Value;
        Inc(P);
        Delta := Delta - dy;
        if Delta > dx then
        begin
          Dec(P, Width);
          Delta := Delta - dx;
        end;
      end;
    end;
  end
  else // dy > Abs(dx)
  begin
    if dx > 0 then Step := 1
    else
    begin
      Step := -1;
      dx := -dx;
    end;
    Delta := dx shr 1;

    x := X1;
    P := PixelPtr[x, Y1];
    for lp1 := Y1 to Y2 do
    begin
      P^ := Value;
      Inc(P, Width);
      Delta := Delta + dx;
      if Delta > dy then
      begin
        Inc(P, Step);
        Delta := Delta - dy;
      end;
    end;

  end;
end;

function TBitmap32.ClipLine(var X0, Y0, X1, Y1: Integer): Boolean;
type
  Edge = (Left, Right, Top, Bottom);
  OutCode = set of Edge;
var
  Accept, AllDone: Boolean;
  OutCode0, OutCode1, OutCodeOut: OutCode;
  X, Y: Integer;

  procedure CompOutCode(X, Y: Integer; var Code: OutCode);
  begin
    Code := [];
    if X < 0 then Code := Code + [Left];
    if X >= Width then Code := Code + [Right];
    if Y < 0 then Code := Code + [Top];
    if Y >= Height then Code := Code + [Bottom];
  end;

begin
  Accept := False;
  AllDone := False;
  CompOutCode(X0, Y0, OutCode0);
  CompOutCode(X1, Y1, OutCode1);
  repeat
    if (OutCode0 = []) and (OutCode1 = []) then // trivial accept and exit
    begin
      Accept := True;
      AllDone := True;
    end
    else if (OutCode0 * OutCode1) <> [] then AllDone := True // trivial reject
    else // calculate intersections
    begin
      if OutCode0 <> [] then OutCodeOut := OutCode0
      else OutCodeOut := OutCode1;
      X := 0;
      Y := 0;
      if Left in OutCodeOut then
        Y := Y0 + (Y1 - Y0) * (-X0) div (X1 - X0)
      else if Right in OutCodeOut then
      begin
        Y := Y0 + (Y1 - Y0) * (Width - 1 - X0) div (X1 - X0);
        X := Width - 1;
      end
      else if Top in OutCodeOut then
        X := X0 + (X1 - X0) * (-Y0) div (Y1 - Y0)
      else if Bottom in OutCodeOut then
      begin
        X := X0 + (X1 - X0) * (Height - 1 - Y0) div (Y1 - Y0);
        Y := Height - 1;
      end;
      if OutCodeOut = OutCode0 then
      begin
        X0 := X;
        Y0 := Y;
        CompOutCode(X0, Y0, OutCode0);
      end
      else
      begin
        X1 := X;
        Y1 := Y;
        CompOutCode(X1, Y1, OutCode1);
      end
    end;
  until AllDone;
  Result := Accept;
end;

function TBitmap32.ClipLineF(var X0, Y0, X1, Y1: Single;
  MinX, MaxX, MinY, MaxY: Single): Boolean;
type
  Edge = (Left, Right, Top, Bottom);
  OutCode = set of Edge;
var
  Accept, AllDone: Boolean;
  OutCode0, OutCode1, OutCodeOut: OutCode;
  X, Y: Single;

  procedure CompOutCode(X, Y: Single; var Code: OutCode);
  begin
    Code := [];
    if X < MinX then Code := Code + [Left];
    if X > MaxX then Code := Code + [Right];
    if Y < MinY then Code := Code + [Top];
    if Y > MaxY then Code := Code + [Bottom];
  end;

begin
  Accept := False; AllDone := False;
  CompOutCode(X0, Y0, OutCode0);
  CompOutCode(X1, Y1, OutCode1);
  repeat
    if (OutCode0 = []) and (OutCode1 = []) then // trivial accept and exit
    begin
      Accept := True;
      AllDone := True;
    end
    else if (OutCode0 * OutCode1) <> [] then AllDone := True // trivial reject
    else // calculate intersections
    begin
      if OutCode0 <> [] then OutCodeOut := OutCode0
      else OutCodeOut := OutCode1;
      X := 0;
      Y := 0;
      if Left in OutCodeOut then
      begin
        Y := Y0 + (Y1 - Y0) * (MinX - X0) / (X1 - X0);
        X := MinX;
      end
      else if Right in OutCodeOut then
      begin
        Y := Y0 + (Y1 - Y0) * (MaxX - X0) / (X1 - X0);
        X := MaxX - 1;
      end
      else if Top in OutCodeOut then
      begin
        X := X0 + (X1 - X0) * (MinY - Y0) / (Y1 - Y0);
        Y := MinY;
      end
      else if Bottom in OutCodeOut then
      begin
        X := X0 + (X1 - X0) * (MaxY - Y0) / (Y1 - Y0);
        Y := MaxY;
      end;
      if OutCodeOut = OutCode0 then
      begin
        X0 := X;
        Y0 := Y;
        CompOutCode(X0, Y0, OutCode0);
      end
      else
      begin
        X1 := X;
        Y1 := Y;
        CompOutCode(X1, Y1, OutCode1);
      end
    end;
  until AllDone;
  Result := Accept;
end;

procedure TBitmap32.DrawLineS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  if ClipLine(X1, Y1, X2, Y2) then DrawLine(X1, Y1, X2, Y2, Value);
end;

procedure TBitmap32.DrawLineT(X1, Y1, X2, Y2: Integer; Value: TColor32);
var
  lp1: Integer;
  x: Integer;
  dy, dx, Step, Delta: Integer;
  P: PColor32;
  a: TColor32;
begin
  a := Value shr 24;
  if a = 0 then Exit;
  if a = $FF then
  begin
    DrawLine(X1, Y1, X2, Y2, Value);
    Exit;
  end;

  if Y1 > Y2 then
  begin
    Swap(X1, X2);
    Swap(Y1, Y2);
  end;

  dx := X2 - X1;
  dy := Y2 - Y1;

  if dy = 0 then
  begin
    TestSwap(X1, X2);
    DrawHorzLineT(X1, Y1, X2, Value);
    Exit;
  end;

  if dx = 0 then
  begin
    DrawVertLineT(X1, Y1, Y2, Value);
    Exit;
  end;

  if Abs(dx) > dy then
  begin
    Delta := dy shr 1;
    if dx >= 0 then
    begin
      P := PixelPtr[X1, Y1];
      for lp1 := X1 to X2 do
      begin
        SET_T(P, a, Value);
        Inc(P);
        Delta := Delta + dy;
        if Delta > dx then
        begin
          Inc(P, Width);
          Delta := Delta - dx;
        end;
      end;
    end
    else // dx < 0
    begin
      dx := -dx;
      dy := -dy;
      P := PixelPtr[X2, Y2];
      for lp1 := X2 to X1 do
      begin
        SET_T(P, a, Value);
        Inc(P);
        Delta := Delta - dy;
        if Delta > dx then
        begin
          Dec(P, Width);
          Delta := Delta - dx;
        end;
      end;
    end;
  end
  else // dy > dx
  begin
    if dx > 0 then Step := 1
    else
    begin
      Step := -1;
      dx := -dx;
    end;
    Delta := dx shr 1;

    x := X1;
    P := PixelPtr[x, Y1];
    for lp1 := Y1 to Y2 do
    begin
      SET_T(P, a, Value);
      Inc(P, Width);
      Delta := Delta + dx;
      if Delta > dy then
      begin
        Inc(P, Step);
        Delta := Delta - dy;
      end;
    end;

  end;
end;

procedure TBitmap32.DrawLineTS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  if ClipLine(X1, Y1, X2, Y2) then DrawLineT(X1, Y1, X2, Y2, Value);
end;

procedure TBitmap32.DrawLineF(X1, Y1, X2, Y2: Single; Value: TColor32);
var
  n, i: Integer;
  px, py, ex, ey, nx, ny, hyp: Integer;
  A: TColor32;
begin
  px := Round(x1 * 65536); py := Round(y1 * 65536);
  ex := Round(x2 * 65536); ey := Round(y2 * 65536);
  nx := ex - px; ny := ey - py;
  hyp := Round(Hypot(nx, ny));
  if hyp < 256 then Exit;
  n := hyp shr 16;
  if n > 0 then
  begin
    nx := Round(nx / hyp * 65536); ny := Round(ny / hyp * 65536);
    for i := 0 to n - 1 do
    begin
      SET_T256(px shr 8, py shr 8, Value);
      px := px + nx;
      py := py + ny;
    end;
  end;
  A := Value shr 24;
  hyp := hyp - n shl 8;
  A := A * Longword(hyp) and $FF000000;
  SET_T256((ex + px - nx) shr 17, (ey + py - ny) shr 17, Value and _RGB + A);
end;

procedure TBitmap32.DrawLineFS(X1, Y1, X2, Y2: Single; Value: TColor32);
var
  n, i: Integer;
  px, py, ex, ey, nx, ny, hyp: Integer;
  A: TColor32;
begin
  if ClipLineF(X1, Y1, X2, Y2, 0, FWidth, 0, FHeight) then
    if (X1 < FWidth - 1) and (X2 < FWidth - 1) and
      (Y1 < FHeight - 1) and (Y2 < FHeight - 1) then
      DrawLineF(X1, Y1, X2, Y2, Value)
    else // check every pixel
    begin
      px := Round(x1 * 65536); py := Round(y1 * 65536);
      ex := Round(x2 * 65536); ey := Round(y2 * 65536);
      nx := ex - px; ny := ey - py;
      hyp := Round(Hypot(nx, ny));
      if hyp < 256 then Exit;
      n := hyp shr 16;
      if n > 0 then
      begin
        nx := Round(nx / hyp * 65536); ny := Round(ny / hyp * 65536);
        for i := 0 to n - 1 do
        begin
          SET_TS256(px shr 8, py shr 8, Value);
          px := px + nx;
          py := py + ny;
        end;
      end;
      A := Value shr 24;
      hyp := hyp - n shl 8;
      A := A * Longword(hyp) and $FF000000;
      SET_TS256((ex + px - nx) shr 17, (ey + py - ny) shr 17, Value and _RGB + A);
    end;
end;

procedure TBitmap32.DrawLineA(X1, Y1, X2, Y2: Integer; Value: TColor32);
var
  DX, DY, S, D: Integer;
  EC, EA: Word;
  CI: Byte;
begin
  if (X1 = X2) or (Y1 = Y2) then
  begin
    DrawLine(X1, Y1, X2, Y2, Value);
    Exit;
  end;

  if Y1 > Y2 then
  begin
    Swap(X1, X2);
    Swap(Y1, Y2);
  end;

  DX := X2 - X1;
  DY := Y2 - Y1;

  if DX > -1 then S := 1
  else
  begin
    S := -1;
    DX := -DX;
  end;

  EC := 0;
  Pixel[X1, Y1] := Value;

  if DY > DX then
  begin
    EA := DX shl 16 div DY;
    while DY > 1 do
    begin
      Dec(DY);
      D := EC;
      Inc(EC, EA);
      if EC <= D then Inc(X1, S);
      Inc(Y1);
      CI := EC shr 8;
      SET_T(@Bits[X1 + Y1 * Width], GAMMA_TABLE[CI xor 255], Value);
      SET_T(@Bits[X1 + S + Y1 * Width], GAMMA_TABLE[CI], Value);
    end;
  end
  else // DY <= DX
  begin
    EA := DY shl 16 div DX;
    while DX > 1 do
    begin
      Dec(DX);
      D := EC;
      Inc(EC, EA);
      if EC <= D then Inc(Y1);
      Inc(X1, S);
      CI := EC shr 8;
      SET_T(@Bits[X1 + Y1 * Width], GAMMA_TABLE[CI xor 255], Value);
      SET_T(@Bits[X1 + (Y1 + 1) * Width], GAMMA_TABLE[CI], Value);
    end;
  end;
end;

procedure TBitmap32.DrawLineAS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  if ClipLine(X1, Y1, X2, Y2) then DrawLineA(X1, Y1, X2, Y2, Value);
end;

procedure TBitmap32.FillRect(X1, Y1, X2, Y2: Integer; Value: TColor32);
var
  j: Integer;
  P: PColor32Array;
begin
  for j := Y1 to Y2 do
  begin
    P := Pointer(GetScanLine(j));
    FillLongword(P[X1], X2 - X1 + 1, Value);
  end;
end;

procedure TBitmap32.FillRectS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  if TestClip(X1, X2, Width) and TestClip(Y1, Y2, Height) then
    FillRect(X1, Y1, X2, Y2, Value);
end;

procedure TBitmap32.FillRectT(X1, Y1, X2, Y2: Integer; Value: TColor32);
var
  i, j: Integer;
  P: PColor32;
  A: Integer;
begin
  A := Value shr 24;
  for j := Y1 to Y2 do
  begin
    P := GetPixelPtr(X1, j);
    for i := X1 to X2 do
    begin
      SET_T(P, A, Value);
      Inc(P);
    end;
  end;
end;

procedure TBitmap32.FillRectTS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  if TestClip(X1, Y1, Width) and TestClip(X2, Y2, Height) then
    FillRectT(X1, Y1, X2, Y2, Value);
end;

procedure TBitmap32.FrameRectS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  DrawHorzLineS(X1, Y1, X2, Value);
  DrawHorzLineS(X1, Y2, X2, Value);
  DrawVertLineS(X1, Y1 + 1, Y2 - 1, Value);
  DrawVertLineS(X2, Y1 + 1, Y2 - 1, Value);
end;

procedure TBitmap32.FrameRectTS(X1, Y1, X2, Y2: Integer; Value: TColor32);
begin
  DrawHorzLineTS(X1, Y1, X2, Value);
  DrawHorzLineTS(X1, Y2, X2, Value);
  DrawVertLineTS(X1, Y1 + 1, Y2 - 1, Value);
  DrawVertLineTS(X2, Y1 + 1, Y2 - 1, Value);
end;

procedure TBitmap32.RaiseRectTS(X1, Y1, X2, Y2: Integer; Contrast: Integer);
var
  C1, C2: TColor32;
begin
  if Contrast > 0 then
  begin
    C1 := clWhite32;
    C2 := clBlack32;
  end
  else if Contrast < 0 then
  begin
    C1 := clBlack32;
    C2 := clWhite32;
    Contrast := -Contrast;
  end
  else Exit;
  Contrast := Contrast * 255 div 100;
  C1 := SetAlpha(C1, Contrast);
  C2 := SetAlpha(C2, Contrast);
  DrawHorzLineTS(X1, Y1, X2 - 1, C1);
  DrawHorzLineTS(X1 + 1, Y2, X2, C2);
  DrawVertLineTS(X1, Y1, Y2 - 1, C1);
  DrawVertLineTS(X2, Y1 + 1, Y2, C2);
end;

procedure TBitmap32.LoadFromStream(Stream: TStream);
var
  B: TBitmap;
begin
  B := TBitmap.Create;
  try
    B.LoadFromStream(Stream);
    Assign(B);
  finally
    B.Free;
  end;
end;

procedure TBitmap32.SaveToStream(Stream: TStream);
var
  B: TBitmap;
begin
  B := TBitmap.Create;
  try
    AssignTo(B);
    B.SaveToStream(Stream);
  finally
    B.Free;
  end;
end;

procedure TBitmap32.DefineProperties(Filer: TFiler);

  function DoWrite: Boolean;
  begin
    if Filer.Ancestor <> nil then
      Result := not (Filer.Ancestor is TGraphic)
    else
      Result := not Empty;
  end;

begin
  Filer.DefineBinaryProperty('Data', ReadData, WriteData, DoWrite);
end;

procedure TBitmap32.ReadData(Stream: TStream);
var
  w, h: Integer;
begin
  Stream.ReadBuffer(w, 4);
  Stream.ReadBuffer(h, 4);
  SetSize(w, h);
  Stream.ReadBuffer(FBits[0], FWidth * FHeight * 4);
end;

procedure TBitmap32.WriteData(Stream: TStream);
begin
  Stream.WriteBuffer(FWidth, 4);
  Stream.WriteBuffer(FHeight, 4);
  Stream.WriteBuffer(FBits[0], FWidth * FHeight * 4);
end;

procedure TBitmap32.LoadFromFile(const FileName: string);
var
  P: TPicture;
begin
  P := TPicture.Create;
  try
    P.LoadFromFile(FileName);
    Assign(P);
  finally
    P.Free;
  end;
end;

procedure TBitmap32.SaveToFile(const FileName: string);
var
  B: TBitmap;
begin
  B := TBitmap.Create;
  try
    AssignTo(B);
    B.SaveToFile(FileName);
  finally
    B.Free;
  end;
end;

procedure TBitmap32.SetFont(Value: TFont32);
begin
  FFont.Assign(Value);
  FontChanged(Self);
end;

procedure TBitmap32.FontChanged(Sender: TObject);
begin
  if FontHandle > 0 then
  begin
    SelectObject(Handle, StockFont);
    DeleteObject(FontHandle);
    FontHandle := 0;
  end;
end;

procedure TBitmap32.UpdateFont;
var
  LFont: TLogFont;
begin
  if FontHandle = 0 then
  begin
    GetObject(Font.Handle, SizeOf(LFont), Addr(LFont));
    LFont.lfEscapement := Font.Escapement;
    FontHandle := CreateFontIndirect(LFont);
    SelectObject(Handle, FontHandle);
    SetTextColor(Handle, ColorToRGB(Font.Color));
    SetBkMode(Handle, Windows.TRANSPARENT);
  end;
end;

procedure TBitmap32.SetSmoothResize(Value: Boolean);
begin
  if FSmoothResize <> Value then
  begin
    FSmoothResize := Value;
    Changed(Self);
  end;
end;

procedure TBitmap32.SetTransparent(const Value: Boolean);
begin
  if FTransparent <> Value then
  begin
    FTransparent := Value;
    if Value then FUseAlpha := False;
    Changed(Self);
  end;
end;

procedure TBitmap32.SetTransparentColor(const Value: TColor32);
begin
  if FTransparentColor <> Value then
  begin
    FTransparentColor := Value;
    Changed(Self);
  end;
end;

procedure TBitmap32.SetMasterAlpha(const Value: Byte);
begin
  if FMasterAlpha <> Value then
  begin
    FMasterAlpha := Value;
    Changed(Self);
  end;
end;

procedure TBitmap32.SetUseAlpha(const Value: Boolean);
begin
  if FUseAlpha <> Value then
  begin
    FUseAlpha := Value;
    if Value then FTransparent := False;
    Changed(Self);
  end;
end;

function TBitmap32.TextExtent(const Text: string): TSize;
begin
  UpdateFont;
  Result.cX := 0;
  Result.cY := 0;
  Windows.GetTextExtentPoint32(Handle, PChar(Text), Length(Text), Result);
end;

procedure TBitmap32.TextOut(X, Y: Integer; const Text: string);
begin
  UpdateFont;
  ExtTextOut(Handle, X, Y, 0, nil, PChar(Text), Length(Text), nil);
end;

procedure TBitmap32.TextOut(X, Y: Integer; const ClipRect: TRect;
  const Text: string);
begin
  UpdateFont;
  ExtTextOut(Handle, X, Y, ETO_CLIPPED, @ClipRect, PChar(Text), Length(Text), nil);
end;

procedure TBitmap32.TextOut(ClipRect: TRect; const Flags: Cardinal;
  const Text: string);
begin
  UpdateFont;
  DrawText(Handle, PChar(Text), Length(Text), ClipRect, Flags);
end;

procedure TBitmap32.SetAutoAlphaMult(const Value: Boolean);
begin
  if FAutoAlphaMult <> Value then
  begin
    FAutoAlphaMult := Value;
    Changed(Self);
  end;
end;

function TBitmap32.TextHeight(const Text: string): Integer;
begin
  Result := TextExtent(Text).cY;
end;

function TBitmap32.TextWidth(const Text: string): Integer;
begin
  Result := TextExtent(Text).cX;
end;

procedure TBitmap32.ColorToGrayscale(ColorSrc: TBitmap32);
var
  I: Integer;
begin
  BeginUpdate;
  try
    if ColorSrc = nil then ColorSrc := Self;
    SetSize(ColorSrc);
    for I := 0 to Width * Height - 1 do
      Bits[I] := Gray32(Intensity(ColorSrc.Bits[I]));
  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.AlphaToGrayscale(AlphaSrc: TBitmap32);
var
  I: Integer;
begin
  BeginUpdate;
  try
    if AlphaSrc = nil then AlphaSrc := Self;
    SetSize(AlphaSrc);
    for I := 0 to Width * Height - 1 do
      Bits[I] := Gray32(AlphaComponent(AlphaSrc.Bits[I]));
  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.IntensityToAlpha(IntensitySrc: TBitmap32);
var
  I: Integer;
  V: TColor32;
begin
  BeginUpdate;
  try
    if IntensitySrc = nil then IntensitySrc := Self;
    SetSize(IntensitySrc);
    for I := 0 to Width * Height - 1 do
    begin
      V := Intensity(IntensitySrc.Bits[I]);
      Bits[I] := Bits[I] and $00FFFFFF + V shl 24;
    end;
  finally
    EndUpdate;
  end;
end;

procedure TBitmap32.Invert(Src: TBitmap32);
var
  I: Integer;
  V: TColor32;
begin
  BeginUpdate;
  try
    if Src = nil then Src := Self;
    SetSize(Src);
    for I := 0 to Width * Height - 1 do
    begin
      V := Src.Bits[I];
      Src.Bits[i] := V xor $00FFFFFF; // do not invert alpha channel
    end;
  finally
    EndUpdate;
  end;
end;


{ TSprite }

procedure TSprite.Changed(Sender: TObject);
begin
  TSpriteSet(Collection).SpriteChanged(Self);
end;

constructor TSprite.Create(Collection: TCollection);
begin
  FBitmap := TBitmap32.Create;
  FBitmap.OnChange := Changed;
  FVisible := True;
  inherited;
end;

destructor TSprite.Destroy;
begin
  FBitmap.Free;
  inherited;
end;

function TSprite.GetPosition: TPoint;
begin
  Result := Point(PositionX, PositionY);
end;

procedure TSprite.SetBitmap(Value: TBitmap32);
begin
  FBitmap.Assign(Value);
end;

procedure TSprite.SetPosition(const Value: TPoint);
begin
  if (FPositionX <> Value.X) or (FPositionY <> Value.Y) then
  begin
    FPositionX := Value.X;
    FPositionY := Value.Y;
    Changed(Self);
  end;
end;

procedure TSprite.SetPositionX(Value: Integer);
begin
  if FPositionX <> Value then
  begin
    FPositionX := Value;
    Changed(Self);
  end;
end;

procedure TSprite.SetPositionY(Value: Integer);
begin
  if FPositionY <> Value then
  begin
    FPositionY := Value;
    Changed(Self);
  end;
end;

procedure TSprite.SetVisible(Value: Boolean);
begin
  if FVisible <> Value then
  begin
    FVisible := Value;
    Changed(Self);
  end;
end;


{ TSpriteSet }

function TSpriteSet.Add: TSprite;
begin
  Result := TSprite(inherited Add);
end;

procedure TSpriteSet.Changed(Sender: TObject);
begin
  if Assigned(FOnChange) then FOnChange(Sender);
end;

constructor TSpriteSet.Create(Image: TImage32);
begin
  inherited Create(TSprite);
  FImage := Image;
end;

function TSpriteSet.GetItem(Index: Integer): TSprite;
begin
  Result := TSprite(inherited GetItem(Index));
end;

function TSpriteSet.GetOwner: TPersistent;
begin
  Result := FImage;
end;

procedure TSpriteSet.SetItem(Index: Integer; Value: TSprite);
begin
  inherited SetItem(Index, Value);
  Changed(Self);
end;

procedure TSpriteSet.SpriteChanged(Sprite: TSprite);
begin
  Changed(Sprite);
end;


{ TImage32 }

procedure TImage32.BitmapChanged(Sender: TObject);
begin
  Invalidate;
end;

constructor TImage32.Create(AOwner: TComponent);
begin
  inherited;
  ControlStyle := [csAcceptsControls, csCaptureMouse, csClickEvents,
    csDoubleClicks, csReplicatable, csOpaque];
  DoubleBuffered := True;
  FBitmap := TBitmap32.Create;
  FBitmap.OnChange := BitmapChanged;
  FSprites := TSpriteSet.Create(Self);
  FSprites.OnChange := SpritesChanged;
  Width := 192;
  Height := 192;
  Scale := 1;
end;

destructor TImage32.Destroy;
begin
  FSprites.Free;
  FBitmap.Free;
  inherited;
end;

procedure TImage32.Paint;
var
  R: TRect;
  W, H, I, Xo, Yo: Integer;
  sx, sy: Single;
begin
  W := Width;
  H := Height;
  R := Rect(0, 0, W, H);
  if csDesigning in ComponentState then Canvas.DrawFocusRect(R);
  if (W = 0) or (H = 0) then Exit;

  Canvas.Lock;
  try

    Bitmap.Lock;
    try
      if not Bitmap.Empty then
      begin
        case ScaleMode of
          smNormal:
            begin
              W := Bitmap.Width;
              H := Bitmap.Height;
            end;

          smStretch:
            begin
              W := Width;
              H := Height;
            end;

          smResize:
            begin
              sx := Width / Bitmap.Width;
              sy := Height / Bitmap.Height;
              if sx >= sy then
              begin
                W := Round(Bitmap.Width * sy);
                H := Height;
              end
              else
              begin
                W := Width;
                H := Round(Bitmap.Height * sx);
              end;
            end;

        else // smScale:
          begin
            W := Round(Bitmap.Width * Scale);
            H := Round(Bitmap.Height * Scale);
          end;

        end;
        R := Rect(0, 0, W, H);
        if BitmapAlign = baCenter then
          OffsetRect(R, (Width - W) div 2, (Height - H) div 2);
      end;

      if Bitmap.Empty or Bitmap.UseAlpha or Bitmap.Transparent or
        (R.Left > 0) or (R.Right < Width) or (R.Top > 0) or (R.Bottom < Height) then
      begin
        Canvas.Brush.Color := Color;
        Canvas.FillRect(Rect(0, 0, Width, Height));
      end;
      Bitmap.Draw(Canvas.Handle, R);
    finally
      Bitmap.Unlock;
    end;

    if Sprites.Count = 0 then Exit;

    { draw sprites }
    Xo := 0;
    Yo := 0;
    if SpritesOrigin = soBitmap then
    begin
      Xo := R.Left;
      Yo := R.Top;
    end;

    for I := 0 to Sprites.Count - 1 do
      with Sprites[I] do
        if Visible then
        begin
          Bitmap.Lock;
          try
            R := Rect(0, 0, Bitmap.Width, Bitmap.Height);
            OffsetRect(R, PositionX + Xo, PositionY + Yo);
            Bitmap.Draw(Canvas.Handle, R);
          finally
            Bitmap.Unlock;
          end;
        end;

  finally
    Canvas.Unlock;
  end;
end;

procedure TImage32.Resize;
begin
  inherited;
  Changed(Self);
end;

procedure TImage32.SetBitmap(Value: TBitmap32);
begin
  FBitmap.Assign(Value);
  Changed(Self);
end;

procedure TImage32.SetBitmapAlign(Value: TBitmapAlign);
begin
  if FBitmapAlign <> Value then
  begin
    FBitmapAlign := Value;
    Changed(Self);
  end;
end;

procedure TImage32.SetScaleMode(Value: TScaleMode);
begin
  if FScaleMode <> Value then
  begin
    FScaleMode := Value;
    Changed(Self);
  end;
end;

procedure TImage32.SetScale(Value: Single);
begin
  if Value < 0.001 then
    raise EPropertyError.Create('Scale should be not less than 0.001');
  if FScale <> Value then
  begin
    FScale := Value;
    Changed(Self);
  end;
end;

procedure TImage32.SetSprites(Value: TSpriteSet);
begin
  FSprites.Assign(Value);
  Changed(Self);
end;

procedure TImage32.SpritesChanged(Sender: TObject);
begin
  Changed(Self);
end;

procedure TImage32.SetSpritesOrigin(Value: TSpriteOrigin);
begin
  if FSpritesOrigin <> Value then
  begin
    FSpritesOrigin := Value;
    Changed(Self);
  end;
end;

procedure TImage32.WMWindowPosChanged(var Message: TWMWindowPosChanged);
begin
  Invalidate;
  inherited;
end;

procedure TImage32.BeginUpdate;
begin
  Inc(UpdateCount);
end;

procedure TImage32.Changed(Sender: TObject);
begin
  if UpdateCount = 0 then Invalidate;
end;

procedure TImage32.EndUpdate;
begin
  Assert(UpdateCount > 0);
  Dec(UpdateCount);
  if UpdateCount = 0 then Changed(Self);
end;

procedure TImage32.SetupBitmap(DoClear: Boolean; ClearColor: TColor32);
begin
  Bitmap.SetSize(Width, Height);
  if DoClear then Bitmap.Clear(ClearColor);
end;

{ TColorManager }

function TColorManager.FindColor(AName: string): TColor32;
var
  I: Integer;
begin
  Result := clBlack32;
  for I := 0 to Count - 1 do
    with TColorEntry(Items[I]^) do
      if Name = AName then
      begin
        Result := Color;
        Break;
      end;
end;

function TColorManager.GetColor(AName: string): TColor32;
var
  S: string;

  function HexToClr(const HexStr: string): Cardinal;
  var
    I: Integer;
    C: Char;
  begin
    Result := 0;
    for I := 1 to Length(HexStr) do
    begin
      C := HexStr[I];
      case C of
        '0'..'9': Result := Int64(16) * Result + (Ord(C) - $30);
        'A'..'F': Result := Int64(16) * Result + (Ord(C) - $37);
        'a'..'f': Result := Int64(16) * Result + (Ord(C) - $57);
      else
        raise EConvertError.Create('Illegal character in hex string');
      end;
    end;
  end;

begin
  S := Trim(AName);
  if S[1] = '$' then S := Copy(S, 2, Length(S) - 1);
  if (S[1] = 'c') and (S[2] = 'l') then Result := FindColor(S)
  else
  try
    Result := HexToClr(S);
  except
    Result := clBlack32;
  end;
end;

function TColorManager.GetColorName(AColor: TColor32): string;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    with TColorEntry(Items[I]^) do
      if Color = AColor then
      begin
        Result := TColorEntry(Items[I]^).Name;
        Exit;
      end;
  Result := '$' + IntToHex(AColor, 8);
end;

destructor TColorManager.Destroy;
var
  I: Integer;
begin
  for I := 0 to Count - 1 do FreeMem(Items[I], SizeOf(TColorEntry));
  inherited;
end;

procedure TColorManager.AddColor(AName: string; AColor: TColor32);
var
  NewEntry: PColorEntry;
begin
  New(NewEntry);
  if NewEntry = nil then
    raise Exception.Create('Could not allocate memory for color registration!');
  with NewEntry^ do
  begin
    Name := AName;
    Color := AColor;
  end;
  Add(NewEntry);
end;

procedure TColorManager.EnumColors(Proc: TGetStrProc);
var
  I: Integer;
begin
  for I := 0 to Count - 1 do Proc(TColorEntry(Items[I]^).Name);
end;

procedure TColorManager.RegisterDefaultColors;
begin
  Capacity := 50;
  AddColor('clBlack32',              clBlack32);
  AddColor('clDimGray32',            clDimGray32);
  AddColor('clGray32',               clGray32);
  AddColor('clLightGray32',          clLightGray32);
  AddColor('clWhite32',              clWhite32);
  AddColor('clMaroon32',             clMaroon32);
  AddColor('clGreen32',              clGreen32);
  AddColor('clOlive32',              clOlive32);
  AddColor('clNavy32',               clNavy32);
  AddColor('clPurple32',             clPurple32);
  AddColor('clTeal32',               clTeal32);
  AddColor('clRed32',                clRed32);
  AddColor('clLime32',               clLime32);
  AddColor('clYellow32',             clYellow32);
  AddColor('clBlue32',               clBlue32);
  AddColor('clFuchsia32',            clFuchsia32);
  AddColor('clAqua32',               clAqua32);

  AddColor('clTrWhite32',            clTrWhite32);
  AddColor('clTrBlack32',            clTrBlack32);
  AddColor('clTrRed32',              clTrRed32);
  AddColor('clTrGreen32',            clTrGreen32);
  AddColor('clTrBlue32',             clTrBlue32);
end;

procedure TColorManager.RemoveColor(AName: string);
var
  I: Integer;
begin
  for I := 0 to Count - 1 do
    if CompareText(TColorEntry(Items[I]^).Name, AName) = 0 then
    begin
      Delete(I);
      Break;
    end;
end;

procedure RegisterColor(AName: string; AColor: TColor32);
begin
  ColorManager.AddColor(AName, AColor);
end;

procedure UnregisterColor(AName: string);
begin
  ColorManager.RemoveColor(AName);
end;


{ Gamma table support for opacities }

procedure SetGamma(Gamma: Single);
var
  i: Integer;
begin
  for i := 0 to 255 do
    GAMMA_TABLE[i] := Round(255 * Power(i / 255, Gamma));
end;


initialization
  SetGamma;
  InitializeCriticalSection(Bitmap32CounterLock);
  StockFont := GetStockObject(SYSTEM_FONT);
  ColorManager := TColorManager.Create;
  ColorManager.RegisterDefaultColors;

finalization
  DeleteCriticalSection(Bitmap32CounterLock);
  ColorManager.Free;

end.

